#include "stdafx.h"

#include <mbstring.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "Util.h"
#include "Logger.h"

////////////////////////////////////////////////////////////////////////
// OǗ
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
// B̃OIuWFNg
//----------------------------------------------------------------------

CLogger CLogger::s_logger;

//----------------------------------------------------------------------
// Ox̕
//----------------------------------------------------------------------

const char* CLogger::s_levelstr[] = {
	"DEBUG",
	"INFO",
	"WARN",
	"ERROR",
	"USER",
	"FATAL",
};

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

CLogger::CLogger() :
	m_logLevel(LOG_DEBUG)
{
	InitializeCriticalSection(&m_csLogger);
}

//----------------------------------------------------------------------
// fXgN^
//----------------------------------------------------------------------

CLogger::~CLogger()
{
	DeleteCriticalSection(&m_csLogger);
}

//----------------------------------------------------------------------
// Oo͂
//----------------------------------------------------------------------

void CLogger::Logging(const LogLevel logLevel, const char* pSourceFilename, const int sourceLineNumber, const char* pFuncname, const std::string* psData, const char* pMessage)
{
	if (logLevel < m_logLevel) {
		return;
	}

	//
	// \[Xt@C̓t@Ĉ݂ɂ
	//
	const char* pFilename = (char*) _mbsrchr((UCHAR*) pSourceFilename, '\\');
	if (pFilename) {
		++pFilename;
	} else {
		pFilename = pSourceFilename;
	}

	//
	// o͂镶𓾂B
	//
	SYSTEMTIME systime;

	GetLocalTime(&systime);

	char message[8192];
	_snprintf(message, sizeof(message), "%04d/%02d/%02d-%02d:%02d:%02d %s: %s: %d: %s: %s\r\n",
				systime.wYear,
				systime.wMonth,
				systime.wDay,
				systime.wHour,
				systime.wMinute,
				systime.wSecond,
				s_levelstr[logLevel],
				pFilename,
				sourceLineNumber,
				pFuncname,
				pMessage
			);
	FORCE_ASCIZ(message);

	//
	// t@C֏o͂
	//
	CSLock cslock(&m_csLogger);

	LogRotate();

	HANDLE hFile = INVALID_HANDLE_VALUE;

	do {
		hFile = CreateFile(m_logFilepath.c_str(), GENERIC_WRITE, 0, 0, OPEN_ALWAYS, 0, 0);
		if (hFile == INVALID_HANDLE_VALUE) {
			break;
		}

		DWORD rc = SetFilePointer(hFile, 0, 0, FILE_END);
		if (rc == INVALID_SET_FILE_POINTER) {
			break;
		}

		DWORD length;
		DWORD writeBytes;

		length = strlen(message);
		WriteFile(hFile, message, length, &writeBytes, 0);

		if (psData && 0 < psData->length()) {

			std::string logdata = *psData;

			// s𓝈ꂷ
			Util::SameCrLf(logdata);

			if (!Util::EndsWith(logdata, "\r\n")) {
				logdata += "\r\n";
			}

			WriteFile(hFile, logdata.c_str(), logdata.length(), &writeBytes, 0);
		}

	} while (false);

	CloseHandle(hFile);
}

//----------------------------------------------------------------------
// O[e[g
//----------------------------------------------------------------------

void CLogger::LogRotate()
{
	struct _stat fsts;

	if (_stat(m_logFilepath.c_str(), &fsts) != 0) {
		return;
	}

	if (!(fsts.st_mode & _S_IFREG)) {
		return;
	}

	if (fsts.st_size <= LOG_FILE_SIZE_MAX) {
		return;
	}

	for (int i = LOG_GENERATION_MAX; 0 < i; --i) {
		char srcFilepath[MAX_PATH * 4];
		char destFilepath[MAX_PATH * 4];

		// t@C
		if (i == 1) {
			_snprintf(srcFilepath, sizeof(srcFilepath), "%s", m_logFilepath.c_str());
		} else {
			_snprintf(srcFilepath, sizeof(srcFilepath), "%s\\%s.%d.%s", m_logFolder.c_str(), LOG_FILE_NAME, i - 1, LOG_FILE_EXT);
		}
		FORCE_ASCIZ(srcFilepath);

		// t@C
		_snprintf(destFilepath, sizeof(destFilepath), "%s\\%s.%d.%s", m_logFolder.c_str(), LOG_FILE_NAME, i, LOG_FILE_EXT);
		FORCE_ASCIZ(destFilepath);

		// t@C폜Ă烊l[
		DeleteFile(destFilepath);
		MoveFile(srcFilepath, destFilepath);
	}
}

//----------------------------------------------------------------------
// 
//----------------------------------------------------------------------
void CLogger::Initialize(const char* pLogFilepath, const char* pLogFolder, const char* pLogLevelStr)
{
	LogLevel logLevel;

	if (_stricmp(pLogLevelStr, "DEBUG") == 0) {
		logLevel = LOG_DEBUG;
	} else if (_stricmp(pLogLevelStr, "INFO") == 0) {
		logLevel = LOG_INFO;
	} else if (_stricmp(pLogLevelStr, "WARN") == 0) {
		logLevel = LOG_WARN;
	} else if (_stricmp(pLogLevelStr, "ERROR") == 0) {
		logLevel = LOG_ERROR;
	} else if (_stricmp(pLogLevelStr, "USER") == 0) {
		logLevel = LOG_USER;
	} else if (_stricmp(pLogLevelStr, "FATAL") == 0) {
		logLevel = LOG_FATAL;
	} else {
		// ftHg
		logLevel = LOG_ERROR;
	}

	s_logger.m_logFilepath = pLogFilepath;
	s_logger.m_logFolder = pLogFolder;
	s_logger.m_logLevel = logLevel;
}

//----------------------------------------------------------------------
// OIuWFNg擾
//----------------------------------------------------------------------

CLogger* CLogger::GetInstance()
{
	return &s_logger;
}

////////////////////////////////////////////////////////////////////////
// Oo̓}NpbpNX
////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------------------
// RXgN^
//----------------------------------------------------------------------

CLogging::CLogging(const CLogger::LogLevel logLevel, const char* pSourceFilename, const int sourceLineNumber, const char* pFuncname) :
	m_logLevel(logLevel),
	m_pSourceFilename(pSourceFilename),
	m_sourceLineNumber(sourceLineNumber),
	m_pFuncname(pFuncname)
{
}

//----------------------------------------------------------------------
// Oo͂
//----------------------------------------------------------------------

void CLogging::operator()(const char* pFormat, ...) const
{
	char message[8192];
	va_list args;

	va_start(args, pFormat);
	_vsnprintf(message, sizeof(message), pFormat, args);
	FORCE_ASCIZ(message);
	va_end(args);

	CLogger::GetInstance()->Logging(m_logLevel, m_pSourceFilename, m_sourceLineNumber, m_pFuncname, 0, message);
}

void CLogging::operator()(const std::string* psData, const char* pFormat, ...) const
{
	char message[8192];
	va_list args;

	va_start(args, pFormat);
	_vsnprintf(message, sizeof(message), pFormat, args);
	FORCE_ASCIZ(message);
	va_end(args);

	CLogger::GetInstance()->Logging(m_logLevel, m_pSourceFilename, m_sourceLineNumber, m_pFuncname, psData, message);
}

void CLogging::operator()(const int length, const char* pData, const char* pFormat, ...) const
{
	char message[8192];
	va_list args;

	va_start(args, pFormat);
	_vsnprintf(message, sizeof(message), pFormat, args);
	FORCE_ASCIZ(message);
	va_end(args);

	std::string sData = std::string(pData, length);

	CLogger::GetInstance()->Logging(m_logLevel, m_pSourceFilename, m_sourceLineNumber, m_pFuncname, &sData, message);
}
